from . import module
import os, sys


class WebAssemblyBinaryLoader:
    def find_module(self, fullname, path=None):
        fname = fullname.split(".")[-1] + ".wasm"
        if path is not None and len(path):
            for p in path:
                if os.path.isfile(os.path.join(p, fname)):
                    self.fname = os.path.join(p, fname)
                    return self

        elif len(fullname.split(".")) < 2 and os.path.isfile(fname):
            self.fname = os.path.abspath(fname)
            return self

    def load_module(self, fullname):
        if fullname in sys.modules:
            return

        with open(self.fname, "rb") as fp:
            mod = module.read_module(fp)
            mod.__file__ = self.fname
            mod.__name__ = fullname
            mod.__loader__ = self

        sys.modules[fullname] = mod
        return mod



from . import sections, values
from .. import native, util
import ctypes, importlib


# extend to add future binary formats
preamble = [b"\0asm\x01\0\0\0"]

sects = [i for j in range(1, 12) for i in (0, j)] + [0]


def read_module(buffer: object) -> dict:
    assert preamble[0] == buffer.read(8), "Invalid magic number or version."

    mod_dict = {
        "custom": (),
        "types": (),
        "imports": (),
        "funcs": [],
        "tables": (),
        "mems": (),
        "globals": (),
        "exports": (),
        "start": None,
        "elem": (),
        "data": (),
    }

    typeidx = None
    code = ()

    # attempt to read sections
    upto = 0
    try:
        while upto < 23:
            id = buffer.read(1)[0]
            assert id < 12
            length = values.read_uint(buffer, 32)

            while sects[upto] != id:
                upto += 1
                if upto == 22:
                    break

            if sects[upto] != id:
                break

            if not id:
                mod_dict["custom"] += (sections.read_customsec(buffer, length),)

            if id == 1:
                mod_dict["types"] = sections.read_typesec(buffer)

            if id == 2:
                mod_dict["imports"] = sections.read_importsec(buffer)

            if id == 3:
                typeidx = sections.read_funcsec(buffer)

            if id == 4:
                mod_dict["tables"] = sections.read_tablesec(buffer)

            if id == 5:
                mod_dict["mems"] = sections.read_memsec(buffer)

            if id == 6:
                mod_dict["globals"] = sections.read_globalsec(buffer)

            if id == 7:
                mod_dict["exports"] = sections.read_exportsec(buffer)

            if id == 8:
                mod_dict["start"] = sections.read_startsec(buffer)

            if id == 9:
                mod_dict["elem"] = sections.read_elemsec(buffer)

            if id == 10:
                code = sections.read_codesec(buffer)

            if id == 11:
                mod_dict["data"] = sections.read_datasec(buffer)

    except IndexError as i:
        pass

    for i in mod_dict["imports"]:
        if i["module"] == "capi":
            mod_dict["funcs"] += (getattr(ctypes.pythonapi, i["name"]),)

        else:
            mod = importlib.import_module(i["module"])
            if i["desc"][0] == "func":
                mod_dict["funcs"] += (
                    {
                        "type": mod_dict["types"][i["desc"][1]],
                        "obj": getattr(mod, i["name"]),
                    },
                )

    for g in mod_dict["globals"]:
        g["offset"] = native.create_global(
            g["globaltype"][0] == "mut", g["globaltype"][1], g["expr"]
        )

    global_offset = native.nativelib.write_globals()
    for g in mod_dict["globals"]:
        g["offset"] += global_offset
        g["obj"] = [
            g["globaltype"][0],
            native.get_global_object(g["offset"], g["globaltype"][1]),
        ]

    mod_dict["funcs"] += [
        native.create_function(
            mod_dict["types"][typeidx[i]][1][0],
            bytes(body),
            bytes(mod_dict["types"][typeidx[i]][0]),
            bytes(locals),
        )
        for i, (locals, body) in enumerate(code)
    ]

    function_base_addr = native.nativelib.write_function_page()
    for funcidx, func in enumerate(mod_dict["funcs"]):
        if isinstance(func, dict):
            obj = ctypes.CFUNCTYPE(func["ret"], *func["params"])(
                function_base_addr + func["address"]
            )
            mod_dict["funcs"][funcidx] = native.wrap_function(
                obj, func["param_clear"]
            )

    native.nativelib.flush_globals()

    for e in mod_dict["exports"]:
        if e["desc"][0] == "func":
            e["obj"] = mod_dict["funcs"][e["desc"][1]]

        elif e["desc"][0] == "global":
            e["obj"] = mod_dict["globals"][e["desc"][1]]["obj"]

    if mod_dict["start"] is not None:
        mod_dict["start"] = mod_dict["funcs"][mod_dict["start"]]

    return util.create_module(mod_dict)



from . import instructions, sections
from .. import native, util
import sexpdata
import ctypes


def read_module(buffer: object) -> dict:
    original = buffer.read()
    buffer.close()
    comment_depth = 0
    stripped = ""
    skip = False
    for i, value in enumerate(original):
        if skip:
            skip = False
            continue

        if i != len(original) - 1:
            if value == "(" and original[i + 1] == ";":
                comment_depth += 1
                skip = True

            elif value == ";" and original[i + 1] == ")":
                comment_depth -= 1
                skip = True

            elif comment_depth == 0:
                stripped += value

        elif comment_depth == 0:
            stripped += value

    try:
        data = sexpdata.loads(stripped)

    except AssertionError:
        # Handle case where (module ...) omitted
        stripped = "(module\n" + stripped + "\n)"
        data = sexpdata.loads(stripped)

    mod_dict = {
        "custom": (),
        "types": [],
        "imports": (),
        "funcs": [],
        "tables": (),
        "mems": (),
        "globals": [],
        "exports": [],
        "start": None,
        "elem": (),
        "data": (),
    }

    type_ids = []
    global_ids = []

    import_funcs = {"funcs": [], "ids": []}
    funcs = {"funcs": [], "ids": []}

    for expr in data:
        if isinstance(expr, list):
            if expr[0].value() == "type":
                typeidx, type = sections.read_type(expr)
                mod_dict["types"].append(type)
                type_ids.append(typeidx)

            elif expr[0].value() == "import":
                type, mod, name, importidx, desc = sections.read_import(expr)
                if type == "func":
                    if mod == "capi":
                        import_funcs["ids"].append(importidx)
                        desc["obj"] = getattr(ctypes.pythonapi, name)
                        import_funcs["funcs"].append(desc)

                    # TODO: .wasm imports

            elif expr[0].value() == "func":
                funcidx, func = sections.read_func(expr)
                funcs["funcs"].append(func)
                funcs["ids"].append(funcidx)

            elif expr[0].value() == "global":
                globalidx, export, global_ = sections.read_global(expr)
                mod_dict["globals"].append(global_)
                global_ids.append(globalidx)
                if export is not None:
                    mod_dict["exports"].append(
                        {
                            "name": export,
                            "type": "global",
                            "idx": len(mod_dict["globals"]) - 1,
                        }
                    )

            elif expr[0].value() == "export":
                mod_dict["exports"].append(sections.read_export(expr))

            elif expr[0].value() == "start":
                if mod_dict["start"] is not None:
                    raise ValueError("Duplicate start symbol")

                try:
                    mod_dict["start"] = expr[1]

                except IndexError:
                    raise ValueError("Invalid start symbol")

    func_ids = import_funcs["ids"] + funcs["ids"]
    mod_dict["funcs"] = import_funcs["funcs"] + funcs["funcs"]

    for g in mod_dict["globals"]:
        for i, term in enumerate(g["init"]):
            if term in global_ids:
                g["init"][i] = global_ids.index(term)

        g["init"] = instructions.read_expr_text(g["init"])
        g["offset"] = native.create_global(
            g["mutable"] == "mut", g["type"], g["init"]
        )

    global_offset = native.nativelib.write_globals()
    for globalidx, g in enumerate(mod_dict["globals"]):
        g["offset"] += global_offset
        g["obj"] = [
            g["mutable"],
            native.get_global_object(g["offset"], g["type"]),
        ]

        mod_dict["globals"][globalidx] = g["obj"]

    for funcidx, func in enumerate(mod_dict["funcs"]):
        if "export" in func and func["export"] is not None:
            mod_dict["exports"].append(
                {
                    "name": func["export"],
                    "type": "func",
                    "idx": funcidx,
                }
            )

        if func["typeidx"] is not None:
            if func["typeidx"] in type_ids:
                func["typeidx"] = type_ids.index(func["typeidx"])

            if func["typeuse"] is not None:
                assert (
                    mod_dict["types"][func["typeidx"]] == func["typeuse"]
                ), "type mismatch"

            func["type"] = mod_dict["types"][func["typeidx"]]

        else:
            if func["typeuse"] is None:
                func["typeuse"] = ([], [])

            func["type"] = func["typeuse"]

        if "obj" not in func:
            for i, term in enumerate(func["body"]):
                if term in global_ids:
                    func["body"][i] = global_ids.index(term)

            func["body"] = instructions.read_expr_text(func["body"])
            if not func["type"][1]:
                func["type"][1].append(0x40)

            func["_obj"] = native.create_function(
                func["type"][1][0],
                bytes(func["body"]),
                bytes(func["type"][0]),
                bytes(func["locals"]),
            )

        else:
            if not hasattr(func["obj"], "func"):
                params, param_clear = native.gen_params(func["type"][0])
                func["obj"].argtypes = params
                if not func["type"][1]:
                    func["obj"].restype = None

                elif func["type"][1][0] == 0x7F:
                    func["obj"].restype = ctypes.c_uint32

                elif func["type"][1][0] == 0x7E:
                    func["obj"].restype = ctypes.c_uint64

                elif func["type"][1][0] == 0x7D:
                    func["obj"].restype = ctypes.c_float

                elif func["type"][1][0] == 0x7C:
                    func["obj"].restype = ctypes.c_double

                func["obj"] = native.wrap_function(
                    func["obj"], param_clear, True
                )

    function_base_addr = native.nativelib.write_function_page()
    for funcidx, func in enumerate(mod_dict["funcs"]):
        if "_obj" in func:
            obj = ctypes.CFUNCTYPE(
                func["_obj"]["ret"], *func["_obj"]["params"]
            )(function_base_addr + func["_obj"]["address"])
            func["obj"] = native.wrap_function(obj, func["_obj"]["param_clear"])

        mod_dict["funcs"][funcidx] = func["obj"]

    if mod_dict["start"] is not None:
        try:
            if mod_dict["start"] in type_ids:
                mod_dict["start"] = mod_dict["funcs"][
                    type_ids.index(mod_dict["start"])
                ]

            else:
                mod_dict["start"] = mod_dict["funcs"][mod_dict["start"]]

        except (IndexError, TypeError):
            raise ValueError("Invalid start symbol")

    native.nativelib.flush_globals()

    for e in mod_dict["exports"]:
        e["name"] = util.sanitize(e["name"])
        if isinstance(e["idx"], sexpdata.Symbol):
            if e["type"] == "func":
                e["idx"] = func_ids.index(e["idx"])

            elif e["type"] == "global":
                e["idx"] = global_ids.index(e["idx"])

        if e["type"] == "func":
            e["obj"] = mod_dict["funcs"][e["idx"]]

        if e["type"] == "global":
            e["obj"] = mod_dict["globals"][e["idx"]]

    return util.create_module(mod_dict)



import sexpdata


valtypes = {
    sexpdata.Symbol("i32"): 0x7F,
    sexpdata.Symbol("i64"): 0x7E,
    sexpdata.Symbol("f32"): 0x7D,
    sexpdata.Symbol("f64"): 0x7C,
}


def flatten(expr: list) -> list:
    if isinstance(expr, (list, tuple)):
        expr = [flatten(sub_expr) for sub_expr in expr]

    return (
        expr.value()
        if hasattr(expr, "value") and not expr.value().startswith("$")
        else expr
    )


def unfold(expr: list) -> list:
    if isinstance(expr, list):
        if len(expr) > 1:
            new_expr = [unfold(i) for i in expr]
            offset = None
            for idx, i in enumerate(new_expr):
                if isinstance(i, list):
                    offset = idx
                    break

            if offset is not None:
                new_expr = [
                    term for sub_expr in new_expr[offset:] for term in sub_expr
                ] + new_expr[:offset]

            return new_expr

        else:
            return unfold(expr[0])

    else:
        return expr


def read_type(expr: list) -> tuple:
    typeidx = None
    if len(expr) == 3:
        typeidx = expr[1]

    assert expr[-1][0].value() == "func", f"Malformed function type '{expr}'"
    t1 = []
    t2 = []
    for sub_expr in expr[-1][1:]:
        if sub_expr[0].value() == "param":
            t1 += flatten(sub_expr[1:])

        if sub_expr[0].value() == "result":
            t2 += flatten(sub_expr[1:])

    for ts in (t1, t2):
        for i, t in enumerate(ts):
            t_sym = sexpdata.Symbol(t)
            if t_sym in valtypes:
                ts[i] = valtypes[t_sym]

    return typeidx, (t1, t2)


def read_typeuse(expr: list) -> tuple:
    typeidx = None

    off = 0
    if (
        off < len(expr)
        and isinstance(expr[off], list)
        and isinstance(expr[off][0], sexpdata.Symbol)
        and expr[off][0].value() == "type"
    ):
        typeidx = expr[off][1]
        off += 1

    params = []
    param_ids = []
    for sub_expr in expr[off:]:
        if (
            isinstance(sub_expr, list)
            and isinstance(sub_expr[0], sexpdata.Symbol)
            and sub_expr[0].value() == "param"
        ):
            if sub_expr[1].value().startswith("$"):
                param_ids.append(sub_expr[1])
                assert sub_expr[2] in valtypes
                params.append(valtypes[sub_expr[2]])

            else:
                for valtype in sub_expr[1:]:
                    assert valtype in valtypes
                    params.append(valtypes[valtype])
                    param_ids.append(None)

            off += 1

    result = None
    if (
        off < len(expr)
        and isinstance(expr[off], list)
        and isinstance(expr[off][0], sexpdata.Symbol)
        and expr[off][0].value() == "result"
    ):
        result = valtypes[expr[off][1]]
        off += 1

    if result is not None:
        typeuse = (params, [result])

    elif params:
        typeuse = (params, [])

    else:
        typeuse = None

    return typeidx, typeuse, param_ids, off


def read_import(expr: list) -> tuple:
    mod = expr[1]
    name = expr[2]

    off = 1

    importidx = None
    if isinstance(expr[3][1], sexpdata.Symbol) and expr[3][
        1
    ].value().startswith("$"):
        importidx = expr[3][1]

    if expr[3][0].value() == "func":
        if importidx is None:
            typeidx, typeuse, _, _ = read_typeuse(expr[3][1:])

        else:
            typeidx, typeuse, _, _ = read_typeuse(expr[3][2:])

        return (
            "func",
            mod,
            name,
            importidx,
            {"typeidx": typeidx, "typeuse": typeuse},
        )


def read_func(expr: list) -> tuple:
    funcidx = None
    export = None
    no_inf = False
    off = 1
    if off < len(expr) and isinstance(expr[off], sexpdata.Symbol):
        if expr[off].value().startswith("$"):
            funcidx = expr[off]
            off += 1

        else:
            no_inf = True

    if not no_inf:
        if (
            off < len(expr)
            and isinstance(expr[off], list)
            and isinstance(expr[off][0], sexpdata.Symbol)
            and expr[off][0].value() == "export"
        ):
            export = expr[off][1]
            off += 1

    typeidx, typeuse, param_ids, o = read_typeuse(expr[off:])
    off += o

    locals = []
    local_ids = []
    if not no_inf:
        while (
            off < len(expr)
            and isinstance(expr[off], list)
            and isinstance(expr[off][0], sexpdata.Symbol)
            and expr[off][0].value() == "local"
        ):
            if (
                len(expr[off]) == 3
                and isinstance(expr[off][1], sexpdata.Symbol)
                and expr[off][1].value().startswith("$")
            ):
                local_ids.append(expr[off][1])
                locals.append(expr[off][2])

            else:
                l = expr[off][1:]
                local_ids += [None for _ in l]
                locals += l

            off += 1

    locals = list(map(lambda x: valtypes[x], locals))
    body = []
    for sub_expr in expr[off:]:
        body.append(flatten(sub_expr))

    if body:
        body = unfold(body)

    # occurs with single instruction without immediates i.e. unreachable, nop
    if not isinstance(body, list):
        body = [body]

    for i, term in enumerate(body):
        if term in param_ids:
            body[i] = param_ids.index(term)

        if term in local_ids:
            body[i] = local_ids.index(term) + len(param_ids)

    return (
        funcidx,
        {
            "export": export,
            "typeidx": typeidx,
            "typeuse": typeuse,
            "locals": locals,
            "body": body,
        },
    )


def read_global(expr: list) -> tuple:
    globalidx = None
    export = None
    globaltype = None
    mutable = "const"
    off = 1
    if isinstance(expr[off], sexpdata.Symbol) and expr[off].value().startswith(
        "$"
    ):
        globalidx = expr[off]
        off += 1

    if isinstance(expr[off], list) and expr[off][0].value() == "export":
        export = expr[off][1]
        off += 1

    if isinstance(expr[off], sexpdata.Symbol):
        globaltype = valtypes[expr[off]]
        off += 1

    elif isinstance(expr[off], list) and expr[off][0].value() == "mut":
        globaltype = valtypes[expr[off][1]]
        mutable = "mut"
        off += 1

    init = expr[off:]
    return (
        globalidx,
        export,
        {"type": globaltype, "mutable": mutable, "init": flatten(init)},
    )


def read_export(expr: list) -> list:
    name = expr[1]
    type = expr[2][0].value()
    idx = expr[2][1]
    return {"name": name, "type": type, "idx": idx}



import json, os, re, types


opcodes = {}
consumes = {}
signatures = {}
with open(os.path.join(os.path.dirname(__file__), "opcodes.json")) as fp:
    data = json.load(fp)
    for group in data["opcodes"]:
        opcodes.update(
            dict(
                zip(
                    (i[0] for i in group["instructions"]),
                    (
                        i + group["offset"]
                        for i in range(len(group["instructions"]))
                    ),
                )
            )
        )

    for k, v in data["consumes"].items():
        consumes[opcodes[k]] = v

    for group in data["opcodes"]:
        signatures.update(
            dict(
                zip(
                    (opcodes[i[0]] for i in group["instructions"]),
                    (i[1:] for i in group["instructions"]),
                )
            )
        )


def create_module(module: dict) -> object:
    """Convert module outlined by dict into WebAssemblyModule object."""
    WebAssemblyModule = type("WebAssemblyModule", (types.ModuleType,), {})
    props = []
    for e in module["exports"]:
        if isinstance(e["obj"], list):

            def get(self, e=e):
                return e["obj"][1][0]

            prop = property(get)

            if e["obj"][0] == "mut":

                def set(self, val, e=e):
                    e["obj"][1][0] = val

                prop = prop.setter(set)

            setattr(WebAssemblyModule, e["name"], prop)

        else:
            setattr(WebAssemblyModule, e["name"], e["obj"])

        props.append(e["name"])

    if module["start"] is not None:
        setattr(WebAssemblyModule, "__call__", module["start"])
        obj = WebAssemblyModule("module")
        obj()

    else:
        obj = WebAssemblyModule("module")

    for prop in props:
        obj.__dict__.update({prop: getattr(obj, prop)})

    obj.__dict__.update({"_internal": module})

    return obj


def expand_bytes(n: int, bits: int = 32) -> list:
    """Convert integer to little endian byte representation."""
    r = []
    for _ in range(bits // 8):
        r.append(n & 255)
        n >>= 8

    return r


def sanitize(name: str) -> str:
    """Convert name to valid Python identifier."""
    if name.isidentifier():
        return name

    if not name[0].isidentifier():
        name = "_" + name

    return name[0] + re.sub(r"\W+", "_", name[1:])



from . import values


type_map = {b"\x7f": "i32", b"\x7e": "i64", b"\x7d": "f32", b"\x7c": "f64"}


def read_valtype(buffer: object) -> str:
    """Read a value type from bufer."""
    try:
        return type_map[buffer.read(1)]

    except KeyError:
        raise TypeError("Invalid value type.")


def read_resulttype(buffer: object) -> str:
    """Read a result type from buffer."""
    if buffer.read(1) == b"\x40":
        # no result type
        return None

    buffer.seek(-1, 1)
    return read_valtype(buffer)


def read_functype(buffer: object) -> tuple:
    """Read a function type from buffer."""
    assert (
        buffer.read(1) == b"\x60"
    ), f"Malformed function type at {buffer.tell() - 1}"
    t1 = [buffer.read(1)[0] for _ in range(values.get_vec_len(buffer))]
    t2 = [buffer.read(1)[0] for _ in range(values.get_vec_len(buffer))]
    if len(t2) == 0:
        t2.append(0x40)

    return t1, t2


def read_limits(buffer: object) -> tuple:
    """Read limits from buffer."""
    try:
        flag = buffer.read(1)[0]
        assert flag in range(2)
        if flag:
            return tuple(values.read_uint(buffer, 32) for _ in range(2))

        else:
            return (values.read_uint(buffer, 32),)

    except (IndexError, AssertionError):
        raise TypeError("Invalid limits.")


def read_memtype(buffer: object) -> tuple:
    """Attempt to read a memory type from buffer."""
    return read_limits(buffer)


def read_tabletype(buffer: object) -> tuple:
    """Attempt to read a table type from buffer."""
    if buffer.read(1) != b"\x70":
        raise TypeError("Invalid element type.")

    return read_limits(buffer)


def read_globaltype(buffer: object) -> tuple:
    """Attempt to read a global type from buffer."""
    t = buffer.read(1)[0]
    try:
        mut = buffer.read(1)[0]
        assert mut in range(2)
        if mut:
            return "mut", t

        return "const", t

    except (IndexError, AssertionError):
        raise TypeError("Invalid global type.")



from . import types, values
from .. import util


def read_expr_binary(buffer: object) -> list:
    """Read an expression from buffer."""
    in_ = []
    instruction = read_instruction_binary(buffer)
    while instruction[0] != "end":
        in_ += instruction
        instruction = read_instruction_binary(buffer)

    return in_


def read_instruction_binary(buffer: object) -> list:
    """Read an instruction from buffer."""
    data = buffer.read(1)[0]

    # Control instructions
    if data in (0, 1):
        return [data]

    if data == 0x2:
        try:
            rt = types.read_valtype(buffer)

        except TypeError:
            buffer.seek(-1, 1)
            if buffer.read(1)[0] != 64:
                raise TypeError("Invalid blocktype.")

            rt = None

        in_ = []
        instruction = read_instruction_binary(buffer)
        while instruction != ["end"]:
            in_.append(instruction)
            instruction = read_instruction_binary(buffer)

        return ["block", rt, tuple(in_), "end"]

    if data == 0x3:
        try:
            rt = types.read_valtype(buffer)

        except TypeError:
            buffer.seek(-1, 1)
            if buffer.read(1)[0] != 64:
                raise TypeError("Invalid blocktype.")

            rt = None

        in_ = []
        instruction = read_instruction_binary(buffer)
        while instruction != "end":
            in_.append(instruction)
            instruction = read_instruction_binary(buffer)

        return ["loop", rt, tuple(in_), "end"]

    if data == 0x4:
        try:
            rt = types.read_valtype(buffer)

        except TypeError:
            buffer.seek(-1, 1)
            if buffer.read(1)[0] != 64:
                raise TypeError("Invalid blocktype.")

            rt = None

        in1 = []
        instruction = read_instruction_binary(buffer)
        while instruction[0] not in ["end", "else"]:
            in1.append(instruction)
            instruction = read_instruction_binary(buffer)

        in1 = tuple(in1)
        if instruction[0] == "end":
            return ["if", rt, in1, "else", (), "end"]

        # "else"
        in2 = []
        instruction = read_instruction_binary(buffer)
        while instruction[0] != "end":
            in2.append(instruction)
            instruction = read_instruction_binary(buffer)

        return ["if", rt, in1, "else", tuple(in2), "end"]

    if data == 0x5:
        return ["else"]

    if data == 0xB:
        return ["end"]

    if data == 0xC:
        return ["br", values.read_uint(buffer, 32)]

    if data == 0xD:
        return ["br_if", values.read_uint(buffer, 32)]

    if data == 0xE:
        l = [
            values.read_uint(buffer, 32)
            for _ in range(values.get_vec_len(buffer))
        ]
        return ["br_table", l, values.read_uint(buffer, 32)]

    if data == 0xF:
        return ["return"]

    if data == 0x10:
        return ["call", values.read_uint(buffer, 32)]

    if data == 0x11:
        x = values.read_uint(buffer, 32)
        assert buffer.read(1)[0] == 0, "Invalid instruction."
        return ["call_indirect", x]

    # Parametric instructions
    if data in (0x1A, 0x1B):
        return [data]

    # Variable instructions
    if data in range(0x20, 0x25):
        return [data] + util.expand_bytes(values.read_uint(buffer, 32))

    # Memory instructions
    if data in range(0x28, 0x3F):
        return (
            [data]
            + util.expand_bytes(values.read_uint(buffer, 32))  # align
            + util.expand_bytes(values.read_uint(buffer, 32))  # offset
        )

    if data in (0x3F, 0x40):
        assert buffer.read(1)[0] == 0
        return [data]

    # Numeric instructions
    if data == 0x41:
        return [data] + util.expand_bytes(values.read_sint(buffer, 32))

    if data == 0x42:
        return [data] + util.expand_bytes(values.read_sint(buffer, 64), 64)

    if data == 0x43:
        return [data] + util.expand_bytes(values.read_f32(buffer))

    if data == 0x44:
        return [data] + util.expand_bytes(values.read_f64(buffer))

    if data in range(0x45, 0xC0):
        return [data]

    raise TypeError("Invalid instruction type.")



from . import instructions, types, values
from .. import util


def read_customsec(buffer: object, length: int) -> tuple:
    """Read a custom section from buffer."""
    start = buffer.tell()
    name = values.read_name(buffer)
    bytes = buffer.read(length - (buffer.tell() - start))
    return name, bytes


def read_typesec(buffer: object) -> list:
    """Read a type section from buffer."""
    return [
        types.read_functype(buffer) for _ in range(values.get_vec_len(buffer))
    ]


def read_importsec(buffer: object) -> tuple:
    """Read an import section from buffer."""
    im = ()
    try:
        for _ in range(values.get_vec_len(buffer)):
            import_ = {
                "module": values.read_name(buffer),
                "name": values.read_name(buffer),
            }
            flag = buffer.read(1)[0]
            assert flag in range(4)
            if not flag:
                import_["desc"] = ("func", values.read_uint(buffer, 32))

            elif flag == 1:
                import_["desc"] = ("table", types.read_tabletype(buffer))

            elif flag == 2:
                import_["desc"] = ("mem", types.read_memtype(buffer))

            elif flag == 3:
                import_["desc"] = (
                    "global",
                    types.types.read_globaltype(buffer),
                )

            im += (import_,)

    except (IndexError, AssertionError):
        raise TypeError("Invalid import section.")

    return im


def read_funcsec(buffer: object) -> tuple:
    """Read a function section from buffer."""
    return tuple(
        values.read_uint(buffer, 32) for _ in range(values.get_vec_len(buffer))
    )


def read_tablesec(buffer: object) -> tuple:
    """Read a table section from buffer."""
    return tuple(
        {"type": types.read_tabletype(buffer)}
        for _ in range(values.get_vec_len(buffer))
    )


def read_memsec(buffer: object) -> tuple:
    """Read a memory section from buffer."""
    return tuple(
        {"type": types.read_memtype(buffer)}
        for _ in range(values.get_vec_len(buffer))
    )


def read_globalsec(buffer: object) -> tuple:
    """Read a global section from buffer."""
    return tuple(
        {
            "globaltype": types.types.read_globaltype(buffer),
            "expr": instructions.read_expr_binary(buffer),
        }
        for _ in range(values.get_vec_len(buffer))
    )


def read_exportsec(buffer: object) -> tuple:
    """Read an export section from buffer."""
    ex = ()
    for _ in range(values.get_vec_len(buffer)):
        export = {"name": util.sanitize(values.read_name(buffer))}
        desc = buffer.read(1)[0]
        assert desc in range(4)
        if not desc:
            export["desc"] = "func", values.read_uint(buffer, 32)

        if desc == 1:
            export["desc"] = "table", values.read_uint(buffer, 32)

        if desc == 2:
            export["desc"] = "mem", values.read_uint(buffer, 32)

        if desc == 3:
            export["desc"] = "global", values.read_uint(buffer, 32)

        ex += (export,)

    return ex


def read_startsec(buffer: object) -> int:
    """Read a start section from buffer."""
    return values.read_uint(buffer, 32)


def read_elemsec(buffer: object) -> tuple:
    """Read an element section from buffer."""
    seg = ()
    for _ in range(values.get_vec_len(buffer)):
        seg += (
            {
                "table": values.read_uint(buffer, 32),
                "offset": instructions.read_expr_binary(buffer),
                "init": tuple(
                    values.read_uint(buffer, 32)
                    for _ in range(values.get_vec_len(buffer))
                ),
            },
        )

    return seg


def read_codesec(buffer: object) -> tuple:
    """Read a code section from buffer."""
    code = ()
    try:
        for _ in range(values.get_vec_len(buffer)):
            size = values.read_uint(buffer, 32)
            start = buffer.tell()

            t = ()
            for _ in range(values.get_vec_len(buffer)):
                n = values.read_uint(buffer, 32)
                v = buffer.read(1)[0]
                t += (tuple(v for _ in range(n)),)

            concat_t = ()
            for locals in t:
                concat_t += locals

            code += ((concat_t, instructions.read_expr_binary(buffer)),)
            end = buffer.tell()
            assert size == end - start

    except AssertionError:
        raise TypeError("Invalid code section.")

    return code


def read_datasec(buffer: object) -> tuple:
    """Read a data section from buffer."""
    return tuple(
        {
            "data": values.read_uint(buffer, 32),
            "offset": instructions.read_expr_binary(buffer),
            "init": tuple(
                buffer.read(1)[0] for _ in range(values.get_vec_len(buffer))
            ),
        }
        for _ in range(values.get_vec_len(buffer))
    )



import math, struct


f32 = struct.Struct("<f")
f64 = struct.Struct("<d")


def get_vec_len(buffer):
    """Return the length of a vector consumed from buffer."""
    return read_uint(buffer, 32)


def read_uint(buffer, bits):
    """Read an unsigned integer stored in LEB128 format."""
    consumed_bytes = 0
    result = 0
    shift = 0

    try:
        while True:
            byte = buffer.read(1)[0]
            consumed_bytes += 1
            assert consumed_bytes <= math.ceil(bits / 7)
            result |= (byte & 127) << shift
            if not byte >> 7:
                break

            shift += 7

    except (AssertionError, IndexError):
        raise TypeError("Invalid integer.")

    return result


def read_sint(buffer, bits):
    """Read a signed integer stored in LEB128 format."""
    try:
        byte = buffer.read(1)[0]
        consumed_bytes = 1
        result = byte & 127
        shift = 7
        while byte >> 7:
            byte = buffer.read(1)[0]
            consumed_bytes += 1
            assert consumed_bytes <= math.ceil(bits / 7)
            result |= (byte & 127) << shift
            shift += 7

    except (AssertionError, IndexError):
        raise TypeError("Invalid integer.")

    if result >> consumed_bytes * 7 - 1:
        result |= ~0 << shift

    return result


def read_f32(buffer):
    """Read a single precision float from buffer."""
    return f32.unpack(buffer.read(4))


def read_f64(buffer):
    """Read a double precision float from buffer."""
    return f64.unpack(buffer.read(8))


def read_name(buffer):
    """Read a name from buffer."""
    return buffer.read(get_vec_len(buffer)).decode("utf-8")



from .. import util
import re, struct


def read_expr_text(source: list) -> list:
    """Read an expression from source."""
    in_ = []
    offset = 0
    while offset < len(source):
        instruction = read_instruction_text(source, offset)
        offset += 1 + instruction.pop()
        in_ += instruction

    return in_


def read_instruction_text(source, offset: int) -> list:
    """Read the instruction from source at offset."""
    data = source[offset]
    if data in util.opcodes:
        op = util.opcodes[data]
        if op not in util.consumes:
            return [op, 0]

        else:
            immediate = source[offset + 1]
            if op in (0x43, 0x44):
                if isinstance(immediate, (float, int)):
                    # 32 bit floating point immediate
                    if op == 0x43:
                        return [op] + list(struct.pack("<f", immediate)) + [1]

                    else:
                        return [op] + list(struct.pack("<d", immediate)) + [1]

                nan = re.match(
                    r"(?P<sign>[+-]?)nan:0x(?P<n>(([0-9a-fA-F]_?)+[0-9a-fA-F]|[0-9a-fA-F]))",
                    immediate,
                )
                if nan is not None:
                    significand = int(nan.group("n"), 16)
                    if op == 0x43:
                        # 32 bit nan
                        if nan.group("sign") not in "+":
                            immediate_value = 0xFF800000

                        else:
                            immediate_value = 0x7F800000

                        assert significand in range(
                            1, 2**23
                        ), "nan significand out of range"

                    else:
                        # 64 bit nan
                        if nan.group("sign") not in "+":
                            immediate_value = 0xFFF0000000000000

                        else:
                            immediate_value = 0x7FF0000000000000

                        assert significand in range(
                            1, 2**52
                        ), "nan significand out of range"

                    immediate_value |= int(nan.group("n"), 16)
                    if op == 0x43:
                        return [op] + util.expand_bytes(immediate_value) + [1]

                    else:
                        return (
                            [op] + util.expand_bytes(immediate_value, 64) + [1]
                        )

                else:
                    hexfloat = re.match(
                        r"(?P<sign>[+-]?)0x(?P<p>(([0-9a-fA-F]_?)+[0-9a-fA-F]|[0-9a-fA-F]))(\.(?P<q>(([0-9a-fA-F]+_?)+[0-9a-fA-F]|[0-9a-fA-F])?))?([pP](?P<e>[+-]?(([0-9]+_?)+[0-9]|[0-9])))?$",
                        immediate,
                    )
                    if hexfloat is not None:
                        sign = 1
                        if hexfloat.group("sign") not in "+":
                            sign = -1

                        value = int(hexfloat.group("p"), 16)
                        for i, digit in enumerate(
                            hexfloat.group("q").replace("_", "")
                        ):
                            value += int(digit, 16) / (16**i)

                        if hexfloat.group("e"):
                            value *= 2 ** int(hexfloat.group("e"))

                        if op == 0x43:
                            return [op] + list(struct.pack("<f", value)) + [1]

                        else:
                            return [op] + list(struct.pack("<d", value)) + [1]

                    else:
                        raise ValueError(
                            f"unrecognized floating point value: {immediate}"
                        )

            sign = 1
            if isinstance(immediate, str) and immediate[0] in ("+", "-"):
                sign, *immediate = immediate
                immediate = "".join(immediate)
                if sign == "-":
                    sign = -1

            if isinstance(immediate, str) and immediate.startswith("0x"):
                immediate = sign * int(immediate[2:], 16)

            return (
                [op] + util.expand_bytes(immediate, util.consumes[op] * 8) + [1]
            )



import glob, json, os, platform, setuptools.command.build_ext, struct, subprocess


opcodes = {}
with open(os.path.join(os.path.dirname(__file__), "wasmpy/opcodes.json")) as fp:
    data = json.load(fp)
    replacements = data["replacements"]
    functions = data["functions"]
    for group in data["opcodes"]:
        opcodes.update(
            dict(
                zip(
                    (i[0] for i in group["instructions"]),
                    (
                        i + group["offset"]
                        for i in range(len(group["instructions"]))
                    ),
                )
            )
        )


def is_x86() -> bool:
    return platform.machine() in ("x86", "i386", "i686", "AMD64", "x86_64")


class assemble(setuptools.Command):
    user_options = [
        (
            "targets=",
            "T",
            "comma separated list of targets to assemble, defaults to current "
            "Python's architecture, e.g. --targets=x86_64,...",
        )
    ]

    def initialize_options(self):
        if is_x86():
            if struct.calcsize("P") == 4:
                self.targets = "x86"

            else:
                self.targets = "x86_64"

        else:
            raise ValueError("Unknown architecture")

    def finalize_options(self):
        self.targets = self.targets.split(",")
        for target in self.targets:
            if target not in os.listdir("wasmpy/arch"):
                raise ValueError(f"Unknown architecture: {target}")

    def log(self, out):
        if self.verbose:
            print(" ".join(out))

    def run(self):
        args = {
            "x86_64": [
                ["as", "-o"],
                ["ld", "--oformat", "binary", "-o"],
                ["ld", "-T", "NUL", "--image-base", "0", "-o"],
                ["objcopy", "-O", "binary", "-j", ".text"],
            ]
        }

        args["x86"] = [i.copy() for i in args["x86_64"]]
        args["x86"][0].insert(1, "--32")
        args["x86"][1].insert(1, "-melf_i386")
        args["x86"][2].insert(1, "-mi386pe")

        for target in self.targets:
            arg = args[target]
            for source in glob.glob(f"wasmpy/arch/{target}/*"):
                if os.path.isdir(source):
                    continue

                if os.path.splitext(source)[1].lower() == ".s":
                    name = os.path.splitext(source)[0]
                    cmd = arg[0] + [name + ".o", source]
                    self.log(cmd)
                    subprocess.call(cmd)

                    if platform.system() == "Linux":
                        cmd = arg[1] + [name, name + ".o"]
                        self.log(cmd)
                        subprocess.call(cmd)

                    elif platform.system() == "Windows":
                        cmd = arg[2] + [name + ".tmp", name + ".o"]
                        self.log(cmd)
                        subprocess.call(cmd)
                        cmd = arg[3] + [name + ".tmp", name]
                        self.log(cmd)
                        subprocess.call(cmd)


class tidy(setuptools.Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        for machine in os.listdir("wasmpy/arch"):
            if os.path.exists(f"wasmpy/arch/{machine}/lib/opcodes.cpp"):
                os.remove(f"wasmpy/arch/{machine}/lib/opcodes.cpp")

            for file in glob.glob(f"wasmpy/arch/{machine}/*"):
                if os.path.isdir(file):
                    continue

                if os.path.splitext(file)[1] != ".s":
                    os.remove(file)


class gen_opcodes(setuptools.Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        # generate opcodes.cpp
        if is_x86():
            if struct.calcsize("P") == 4:
                machine = "x86"

            else:
                machine = "x86_64"

        else:
            raise ValueError("Unknown architecture")

        with open(f"wasmpy/arch/{machine}/lib/opcodes.cpp", "w+") as out:
            out.writelines(
                (
                    "// auto-generated\n\n",
                    '#include "opcodes.hpp"\n\n',
                    "bool decodeOperation(bytes buf, size_t offset, bytes *insts)\n{\n\t",
                    "switch (buf.at(offset))\n\t{\n\t",
                )
            )

            for file in glob.glob(f"wasmpy/arch/{machine}/*"):
                if os.path.isdir(file):
                    continue

                inst = os.path.basename(file)

                if inst in opcodes:
                    with open(file, "rb") as fp:
                        data = ", ".join(str(i) for i in fp.read() if i != 0x90)

                    # apply replacements
                    if inst in replacements:
                        for replacement in replacements[inst]:
                            data = data.replace(*replacement)

                    out.write(f"case {opcodes[inst]}:\n\t\t")

                    out.write(f"*insts = {{{data}}};\n\t\t")

                    out.write("break;\n\n\t")

            out.write(
                'default:\n\t\tPyErr_SetString(PyExc_NotImplementedError, "unimplemented instruction");\n\t\treturn false;\n\t}\n\treturn true;\n}\n'
            )

            for file in glob.glob(f"wasmpy/arch/{machine}/*"):
                if os.path.isdir(file):
                    continue

                name = os.path.basename(file)
                if name not in opcodes:
                    with open(file, "rb") as fp:
                        data = ", ".join(str(i) for i in fp.read() if i != 0x90)

                    if os.path.splitext(name)[1] == "":
                        # apply replacements
                        if name in replacements:
                            for replacement in replacements[name]:
                                data = data.replace(*replacement)

                        if name in functions:
                            out.write(
                                f"\nbytes {name}({functions[name]})\n{{\n\treturn {{{data}}};\n}}\n"
                            )

                        else:
                            out.write(f"\nbytes {name} = {{{data}}};\n")


class build_ext(setuptools.command.build_ext.build_ext):
    def run(self):
        self.run_command("gen_opcodes")
        setuptools.command.build_ext.build_ext.run(self)


ext = [
    setuptools.Extension(
        "wasmpy.wasi_unstable",
        sources=["wasmpy/wasi.c"],
        define_macros=[("WASI_UNSTABLE", None)],
        py_limited_api=True,
    ),
    setuptools.Extension(
        "wasmpy.wasi_snapshot_preview1",
        sources=["wasmpy/wasi.c"],
        py_limited_api=True,
    ),
]

if is_x86():
    if struct.calcsize("P") == 4:
        machine = "x86"

    else:
        machine = "x86_64"

    ext.append(
        setuptools.Extension(
            "wasmpy.nativelib",
            include_dirs=[
                os.path.abspath("wasmpy"),
                os.path.abspath(f"wasmpy/arch/{machine}/lib"),
            ],
            sources=[
                "wasmpy/nativelib.cpp",
                f"wasmpy/arch/{machine}/lib/lib.cpp",
                f"wasmpy/arch/{machine}/lib/opcodes.cpp",
            ],
            py_limited_api=True,
        )
    )


with open("README.md", "r") as fp:
    description = fp.read()

setuptools.setup(
    name="wasmpy",
    version="0.2.0",
    author="Olivia Ryan",
    author_email="olivia.r.dev@gmail.com",
    description="WebAssembly from Python.",
    long_description=description,
    long_description_content_type="text/markdown",
    url="https://github.com/olivi-r/wasmpy",
    packages=["wasmpy", "wasmpy.binary", "wasmpy.text"],
    package_data={"wasmpy": ["opcodes.json"]},
    ext_modules=ext,
    options={"bdist_wheel": {"py_limited_api": "cp36"}},
    cmdclass={
        "assemble": assemble,
        "tidy": tidy,
        "build_ext": build_ext,
        "gen_opcodes": gen_opcodes,
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Environment :: WebAssembly",
        "Programming Language :: C++",
        "Programming Language :: Assembly",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
    ],
    license="MIT",
    python_requires=">=3.6",
    install_requires=["sexpdata"],
)



from . import nativelib, util
import ctypes


class ResultVoid(ctypes.Structure):
    _fields_ = [("errno", ctypes.c_byte)]
    _pack_ = 1


class ResultI32(ctypes.Structure):
    _fields_ = [("errno", ctypes.c_byte), ("value", ctypes.c_int32)]
    _pack_ = 1


class ResultF32(ctypes.Structure):
    _fields_ = [("errno", ctypes.c_byte), ("value", ctypes.c_float)]
    _pack_ = 1


class ResultI64(ctypes.Structure):
    _fields_ = [("errno", ctypes.c_byte), ("value", ctypes.c_int64)]
    _pack_ = 1


class ResultF64(ctypes.Structure):
    _fields_ = [("errno", ctypes.c_byte), ("value", ctypes.c_double)]
    _pack_ = 1


def create_global(mut, globaltype, expr):
    assert len(expr), "Missing initializer"
    ret, address = create_function(globaltype, bytes(expr), standalone=True)
    value = ctypes.CFUNCTYPE(ret)(address)()

    if isinstance(value.contents, ResultF32):
        value = ctypes.cast(value, ctypes.POINTER(ResultI32))

    elif isinstance(value.contents, ResultF64):
        value = ctypes.cast(value, ctypes.POINTER(ResultI64))

    return nativelib.append_global(value.contents.value, mut, globaltype)


def get_global_object(offset, globaltype):
    if globaltype == 0x7F:
        return ctypes.cast(offset, ctypes.POINTER(ctypes.c_int32))

    if globaltype == 0x7E:
        return ctypes.cast(offset, ctypes.POINTER(ctypes.c_int64))

    if globaltype == 0x7D:
        return ctypes.cast(offset, ctypes.POINTER(ctypes.c_float))

    if globaltype == 0x7C:
        return ctypes.cast(offset, ctypes.POINTER(ctypes.c_double))


def create_function(ret, code, arg=b"", local=b"", standalone=False):
    address = nativelib.create_function(
        ret, code, arg, local, util.consumes, util.signatures, standalone
    )

    params, param_clear = gen_params(arg)

    if ret == 0x7F:
        ret = ctypes.POINTER(ResultI32)

    elif ret == 0x7E:
        ret = ctypes.POINTER(ResultI64)

    elif ret == 0x7D:
        ret = ctypes.POINTER(ResultF32)

    elif ret == 0x7C:
        ret = ctypes.POINTER(ResultF64)

    else:
        ret = ctypes.POINTER(ResultVoid)

    if standalone:
        return ret, address

    return {
        "ret": ret,
        "params": params,
        "address": address,
        "param_clear": param_clear,
    }


def gen_params(arg):
    params = []
    param_clear = []
    for i, a in enumerate(arg):
        param_clear.append(f"p{i}")
        if a == 0x7F:
            params.append(ctypes.c_uint32)

        if a == 0x7E:
            params.append(ctypes.c_uint64)

        if a == 0x7D:
            params.append(ctypes.c_float)

        if a == 0x7C:
            params.append(ctypes.c_double)

    return params, param_clear


def wrap_function(func, param_clear, raw=False):
    if raw:
        wrapper = eval(
            f"lambda {', '.join(['self'] + param_clear)}: func({', '.join(param_clear)})",
            {"func": func},
        )

    else:
        # ensure result structs
        wrapper = eval(
            f"lambda {', '.join(['self'] + param_clear)}: ensure(func({', '.join(param_clear)}))",
            {"ensure": ensure, "func": func},
        )

    wrapper.addr = ctypes.cast(func, ctypes.c_void_p).value
    return wrapper


def ensure(result):
    if result.contents.errno == 0:
        if not isinstance(result.contents, ResultVoid):
            return result.contents.value

    elif result.contents.errno == 1:
        raise RuntimeError("unreachable")

    elif result.contents.errno == 2:
        raise ZeroDivisionError("division by zero")

    elif result.contents.errno == 3:
        raise RuntimeError("division overflow")

    elif result.contents.errno == 4:
        raise ZeroDivisionError("integer modulo by zero")

    elif result.contents.errno == 5:
        raise FloatingPointError("unrepresentable truncation result")

    elif result.contents.errno == 6:
        raise NotImplementedError("unimplemented instruction")
